
// GripFit
// Avisynth filter to resize while preserving aspect ratio
//
// By Ross Thomas <ross@grinfinity.com>
//

// This obsolete read the license.txt
// There is no copyright on this code, and there are no conditions
// on its distribution or use. Do with it what you will.


// Modified by Lauri Hahne <lhahne@raa.to> to work with
// Avisynth 2.5 (Beta)
// Copyright 2003 Lauri Hahne

#include <assert.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>

#include <map>
using namespace std;

#include "avisynth25.h"

class Dimensions
{
public:
	Dimensions(int w, int h) : width(w), height(h) {}
	Dimensions(AVSValue w, int h) : width(w), height(h) {}

public:
	AVSValue width;
	int height;
};

class Borders
{
public:
	Borders(int l, int t, int r, int b) : left(l), top(t), right(r), bottom(b) {}

public:
	int left, top, right, bottom;
};

class GripCrop : public GenericVideoFilter
{
public:
	GripCrop(PClip child, AVSValue width, int height, AVSValue source_anamorphic,
		AVSValue dest_anamorphic, int overscan,
		int crop_round_width, int crop_round_height, int resize_round_width,
		int resize_round_height, int luma_threshold, int samples,
		IScriptEnvironment* env);
	virtual PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);

private:
	enum STANDARD { UNKNOWN, NTSC, PAL };

private:
	STANDARD DetermineStandard(const Dimensions& dim);
	double DeterminePAR(STANDARD standard, const Dimensions& dim);
	bool IsAnamorphic(STANDARD standard, const Dimensions& dim, AVSValue def);
	Borders FindSourceBorders(int luma_threshold, int samples, IScriptEnvironment* env);
	void ExamineFrame(int n, int luma_threshold,
		int& left, int& top, int& right, int& bottom, IScriptEnvironment* env);
	int MostCommon(int* vals, int len);
	Dimensions HandleOverscan(const Dimensions& dim, int overscan);
	Dimensions CalculateResize(const Dimensions& source_dim, Borders& source_borders,
		double source_par, bool source_is_anamorphic, const Dimensions& dest_dim,
		double dest_par,bool dest_is_anamorphic, int crop_round_width, int crop_round_height,
		int resize_round_width, int resize_round_height);

private:
	PClip crop_clip;
};

GripCrop::GripCrop(PClip _child, AVSValue width, int height, AVSValue source_anamorphic,
				   AVSValue dest_anamorphic, int overscan,
				   int crop_round_width, int crop_round_height, int resize_round_width,
				   int resize_round_height, int luma_threshold, int samples,
				   IScriptEnvironment* env)
	: GenericVideoFilter(_child)
{
	// Get source and dest information

	Dimensions source_dim(vi.width, vi.height), dest_dim(width, height);

	STANDARD source_standard = DetermineStandard(source_dim),
		dest_standard = DetermineStandard(dest_dim);

	double fps = double(vi.fps_numerator) / vi.fps_denominator;
	if(NTSC == source_standard && fps != 29.97 && fps != 23.976)
		source_standard = UNKNOWN;
	else if(PAL == source_standard && fps != 25.0)
		source_standard = UNKNOWN;

	double source_par = DeterminePAR(source_standard, source_dim), dest_par;
	if(0.0 == source_par)
		source_par = 1.0; // Assume square pixels if unrecognized
	if(dest_dim.width.Defined())
	{
		dest_par = DeterminePAR(dest_standard, dest_dim);
		if(0.0 == dest_par)
			env->ThrowError("GripCrop: Unrecognized destination frame size");
	} else
		dest_par = 1.0; // Assume square pixels when only height is defined

	bool source_is_anamorphic = IsAnamorphic(source_standard, source_dim, source_anamorphic),
		dest_is_anamorphic = IsAnamorphic(dest_standard, dest_dim, dest_anamorphic);

	// Compensate for overscan blocks

	dest_dim = HandleOverscan(dest_dim, overscan);

	// Find existing borders in source

	Borders source_borders = FindSourceBorders(luma_threshold, samples, env);

	// Calculate resize values

	Dimensions resize_dim = CalculateResize(source_dim, source_borders, source_par,
		source_is_anamorphic, dest_dim, dest_par, dest_is_anamorphic, crop_round_width,
		crop_round_height, resize_round_width, resize_round_height);

	// Build filter chain

	vi.width = source_dim.width.AsInt() - source_borders.left - source_borders.right;
	vi.height = source_dim.height - source_borders.top - source_borders.bottom;

	AVSValue crop_args[5] = { child, source_borders.left, source_borders.top, vi.width, vi.height };
	crop_clip = env->Invoke("Crop", AVSValue(crop_args, 5)).AsClip();

	// Save variables for later use

	if(!env->SetVar("GripFit_target_width", width) ||
	   !env->SetVar("GripFit_target_height", AVSValue(height)) ||
	   !env->SetVar("GripFit_resize_width", resize_dim.width) ||
	   !env->SetVar("GripFit_resize_height", AVSValue(resize_dim.height)))
	{
		env->ThrowError("GripCrop: Could not save target frame size as variables");
	}
}

PVideoFrame __stdcall GripCrop::GetFrame(int n, IScriptEnvironment* env)
{
	return crop_clip->GetFrame(n, env);
}

GripCrop::STANDARD GripCrop::DetermineStandard(const Dimensions& dim)
{
	if(240 == dim.height || 480 == dim.height)
		return NTSC;
	else if(288 == dim.height || 576 == dim.height)
		return PAL;
	else
		return UNKNOWN;
}

double GripCrop::DeterminePAR(GripCrop::STANDARD standard, const Dimensions& dim)
{
	int w = dim.width.AsInt(), h = dim.height;

	if(NTSC == standard)
	{
		if(720 == w && 480 == h) // DVD
			return 72.0 / 79.0;
		else if(704 == w && 480 == h) // KVCDx2
			return 72.0 / 79.0; // FIXME?
		else if(544 == w && 480 == h) // KVCDx3
			return (72.0 / 79.0) / (3.0 / 4.0);
		else if(528 == w && 480 == h) // KVCDx3
			return (72.0 / 79.0) / (3.0 / 4.0);
		else if(480 == w && 480 == h) // SVCD
			return 108.0 / 79.0;
		else if(352 == w && 480 == h) // 1/2 D1
			return 144.0 / 79.0;
		else if(352 == w && 240 == h) // VCD
			return 72.0 / 79.0;
	} else if(PAL == standard)
	{
		if(720 == w && 576 == h) // DVD
			return 128.0 / 117.0;
		else if(704 == w && 576 == h) // KVCDx2
			return 128.0 / 117.0; // FIXME?
		else if(544 == w && 576 == h) // KVCDx3
			return (128.0 / 117.0) / 0.75;
		else if(528 == w && 576 == h) // KVCDx3
			return (128.0 / 117.0) / 0.75;
		else if(480 == w && 576 == h) // SVCD
			return 128.0 / 78.0;
		else if(352 == w && 576 == h) // 1/2 D1
			return 256.0 / 117.0;
		else if(352 == w && 288 == h) // VCD
			return 128.0 / 117.0;
	}

	// Unknown

	return 0.0;
}

bool GripCrop::IsAnamorphic(GripCrop::STANDARD standard, const Dimensions& dim, AVSValue def)
{
	if(def.Defined())
		return def.AsBool();
	if(!dim.width.Defined())
		return false;

	int w = dim.width.AsInt(), h = dim.height;
	if(NTSC == standard && 720 == w && 480 == h)
		return true;
	if(PAL == standard && 720 == w && 576 == h)
		return true;

	return false;
}

Dimensions GripCrop::HandleOverscan(const Dimensions& dim, int overscan)
{
	if(!dim.width.Defined())
		return dim;
	overscan *= 16;
	return Dimensions(dim.width.AsInt() - overscan, dim.height - overscan);
}

Borders GripCrop::FindSourceBorders(int luma_threshold, int samples, IScriptEnvironment* env)
{
	int *lefts = new int[samples],
		*tops = new int[samples],
		*rights = new int[samples],
		*bottoms = new int[samples];

	// Examine sample frames

	int skip = vi.num_frames / (samples + 1), n = skip >> 1;
	for(int i = 0; i < samples; ++i)
	{
		ExamineFrame(n, luma_threshold, lefts[i], tops[i], rights[i], bottoms[i], env);
		n += skip;
		assert(n < vi.num_frames);
	}

	// Find most common value for each border

	int left = MostCommon(lefts, samples),
		top = MostCommon(tops, samples),
		right = MostCommon(rights, samples),
		bottom = MostCommon(bottoms, samples);

	delete [] lefts;
	delete [] tops;
	delete [] rights;
	delete [] bottoms;

	// Sanity check

	if(left + right > vi.width / 2 || top + bottom > vi.height / 2)
		env->ThrowError("GripCrop: Auto-detection of borders failed");

	return Borders(left, top, right, bottom);
}

void GripCrop::ExamineFrame(int n, int luma_threshold,
							int& left, int& top, int& right, int& bottom,
							IScriptEnvironment* env)
{
	PVideoFrame f = child->GetFrame(n, env);
	const BYTE* p = f->GetReadPtr();
	const BYTE* pp = p;
	const int height = f->GetHeight(), pitch = f->GetPitch(), row_size = f->GetRowSize();

	left = top = right = bottom = 0;

	for(int y = 0; y < height; ++y)
	{
		for(int x = 0; x < row_size; x += 2)
		{
			if(p[x] > luma_threshold)
			{
				top = y;
				goto do_bottom;
			}
		}
		p += pitch;
	}

do_bottom:

	p = pp + (height - 1) * pitch;
	for(int y1 = 0; y1 < height; ++y1)
	{
		for(int x = 0; x < row_size; x += 2)
		{
			if(p[x] > luma_threshold)
			{
				bottom = y1;
				goto do_left;
			}
		}
		p -= pitch;
	}

do_left:

	for(int x = 0; x < row_size; x += 2)
	{
		p = pp;
		for(int y = 0; y < height; ++y)
		{
			if(p[x] > luma_threshold)
			{
				left = x >> 1;
				goto do_right;
			}
			p += pitch;
		}
	}

do_right:

	for(int x1 = row_size - 2; x1 >= 0; x1 -= 2)
	{
		p = pp;
		for(int y = 0; y < height; ++y)
		{
			if(p[x] > luma_threshold)
			{
				right = ((row_size - 2) - x1) >> 1;
				return;
			}
			p += pitch;
		}
	}
}

int GripCrop::MostCommon(int* vals, int len)
{
	map<int, int> m;
	for(int i = 0; i < len; ++i)
	{
		map<int, int>::iterator it = m.find(vals[i]);
		if(it == m.end())
			m[vals[i]] = 1;
		else
			++m[vals[i]];
	}

	map<int, int>::iterator most = m.begin(), it = most;
	for(++it; it != m.end(); ++it)
	{
		if(it->second > most->second)
			most = it;
	}

	return most->first;
}

Dimensions GripCrop::CalculateResize(const Dimensions& source_dim, Borders& source_borders,
									 double source_par, bool source_is_anamorphic,
									 const Dimensions& dest_dim, double dest_par,
									 bool dest_is_anamorphic, int crop_round_width,
									 int crop_round_height, int resize_round_width,
									 int resize_round_height)
{
	int w = 0, h = 0;
	double hscale = 0.0, vscale = 0.0;

	// First, try resizing by width (if we can)

	if(dest_dim.width.Defined())
	{
		hscale = (dest_dim.height / dest_par) / (source_dim.height / source_par);
		vscale = (dest_dim.width.AsInt() * dest_par) /
			(source_dim.width.AsInt() * source_par);
		w = dest_dim.width.AsInt();
		h = int((source_dim.height - source_borders.top -
			source_borders.bottom) * vscale + 0.5);
		if(h > dest_dim.height)
			h = 0;
	}

	// If that doesn't work, resize by height

	if(0 == h)
	{
		hscale = (dest_dim.height / dest_par) / (source_dim.height / source_par);
		w = int((source_dim.width.AsInt() - source_borders.left -
			source_borders.right) * hscale + 0.5);
		if(dest_dim.width.Defined())
		{
			vscale = (dest_dim.width.AsInt() * dest_par) /
				(source_dim.width.AsInt() * source_par);
		} else
			vscale = (w * dest_par) / (source_dim.width.AsInt() * source_par);
		h = dest_dim.height;
	}

	// Account for anamorphism

	if(source_is_anamorphic && !dest_is_anamorphic)
	{
		if(dest_dim.width.Defined())
			h = int(h * 0.75);
		else
			w = int(w / 0.75);
	} else if(!source_is_anamorphic && dest_is_anamorphic)
	{
		if(dest_dim.width.Defined())
			h = int(h / 0.75);
		else
			w = int(w * 0.75);
	}

	// Round

	int rw = (w + resize_round_width - 1) / resize_round_width * resize_round_width,
		rh = (h + resize_round_height - 1) / resize_round_height * resize_round_height;

	if((dest_dim.width.Defined() && rw > dest_dim.width.AsInt()) || (rh > dest_dim.height))
	{
        rw = w / resize_round_width * resize_round_width;
		rh = h / resize_round_height * resize_round_height;
	}

	// Calculate cropping adjustments

	double real_dar = double(w) / h;
	int real_width = int(rh * real_dar), cw = 0, ch = 0;

	if(rw > real_width) // Compensate by cropping vertically
		ch = int((rw - real_width) * real_dar / vscale);
	else if(rw < real_width) // Compensate by cropping horizontally
		cw = int((real_width - rw) / hscale);

	source_borders.left += cw / 2;
	source_borders.top += ch / 2;
	source_borders.right += cw - cw / 2;
	source_borders.bottom += ch - ch / 2;

	// Round down

	cw = source_dim.width.AsInt() - source_borders.left - source_borders.right;
	ch = source_dim.height - source_borders.top - source_borders.bottom;
	int rcw = cw / crop_round_width * crop_round_width,
		rch = ch / crop_round_height * crop_round_height;
	cw -= rcw;
	ch -= rch;
	source_borders.left += cw / 2;
	source_borders.top += ch / 2;
	source_borders.right += cw - cw / 2;
	source_borders.bottom += ch - ch / 2;

	// Ensure crop values are even

	if(source_borders.left % 2)
		++source_borders.left;
	if(source_borders.top % 2)
		++source_borders.top;
	if(source_borders.right % 2)
		++source_borders.right;
	if(source_borders.bottom % 2)
		++source_borders.bottom;

	// Done

	return Dimensions(rw, rh);
}

AVSValue __cdecl Create_GripCrop(AVSValue args, void* /* user_data */,
								 IScriptEnvironment* env)
{
	enum ARGS
	{
		CLIP, WIDTH, HEIGHT, SOURCE_ANAMORPHIC, DEST_ANAMORPHIC, OVERSCAN,
		CROP_ROUND_WIDTH, CROP_ROUND_HEIGHT, RESIZE_ROUND_WIDTH, RESIZE_ROUND_HEIGHT,
		LUMA_THRESHOLD, SAMPLES
	};

	if(!args[CLIP].AsClip()->GetVideoInfo().IsYUV())
		env->ThrowError("GripCrop: Source must be YUY2");

	if(!args[HEIGHT].Defined())
		env->ThrowError("GripCrop: Must specify a height");

	AVSValue width = args[WIDTH];
	int height = args[HEIGHT].AsInt();
	AVSValue source_anamorphic = args[SOURCE_ANAMORPHIC],
		dest_anamorphic = args[DEST_ANAMORPHIC];
	int overscan = args[OVERSCAN].AsInt(0);
	int crop_round_width = args[CROP_ROUND_WIDTH].AsInt(4),
		crop_round_height = args[CROP_ROUND_HEIGHT].AsInt(4),
		resize_round_width = args[RESIZE_ROUND_WIDTH].AsInt(16),
		resize_round_height = args[RESIZE_ROUND_HEIGHT].AsInt(16),
		luma_threshold = args[LUMA_THRESHOLD].AsInt(60),
		samples = args[SAMPLES].AsInt(5);

	if(width.Defined() && width.AsInt() < 1)
		env->ThrowError("GripCrop: width must be > 0");
	if(height < 1)
		env->ThrowError("GripCrop: height must be > 0");
	if(overscan < 0)
		env->ThrowError("GripCrop: overscan must be >= 0");
	if(crop_round_width < 1)
		env->ThrowError("GripCrop: crop_round_width must be > 0");
	if(crop_round_height < 1)
		env->ThrowError("GripCrop: crop_round_height must be > 0");
	if(resize_round_width < 1)
		env->ThrowError("GripCrop: resize_round_width must be > 0");
	if(resize_round_height < 1)
		env->ThrowError("GripCrop: resize_round_height must be > 0");
	if(luma_threshold < 0 || luma_threshold > 255)
		env->ThrowError("GripCrop: luma_threshold must be >= 0 and <= 255");
	if(samples < 1)
		env->ThrowError("GripCrop: samples must be > 0");

	return new GripCrop(args[CLIP].AsClip(), width, height, source_anamorphic,
		dest_anamorphic, overscan, crop_round_width, crop_round_height,
		resize_round_width, resize_round_height, luma_threshold, samples, env);
}

class GripSize : public GenericVideoFilter
{
public:
	GripSize(PClip child, int w, int h, const char* resizer, IScriptEnvironment* env);
	virtual PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);

private:
	PClip resize_clip;
};

GripSize::GripSize(PClip _child, int w, int h, const char* resizer, IScriptEnvironment* env)
	: GenericVideoFilter(_child)
{
	AVSValue args[] = { child, w, h };
	resize_clip = env->Invoke(resizer, AVSValue(args, 3)).AsClip();
	vi.width = w;
	vi.height = h;
}

PVideoFrame __stdcall GripSize::GetFrame(int n, IScriptEnvironment* env)
{
	return resize_clip->GetFrame(n, env);
}

AVSValue __cdecl Create_GripSize(AVSValue args, void* /* user_data */,
								 IScriptEnvironment* env)
{
	enum ARGS { CLIP, RESIZER };

	const char* resizer = args[RESIZER].AsString("BilinearResize");
	if(!env->FunctionExists(resizer))
		env->ThrowError("GripSize: Specified resizer does not exist");

	AVSValue w = env->GetVar("GripFit_resize_width"),
		h = env->GetVar("GripFit_resize_height");
	if(!w.Defined() || !h.Defined())
		env->ThrowError("GripSize: Must run GripCrop before GripSize");

	return new GripSize(args[CLIP].AsClip(), w.AsInt(), h.AsInt(), resizer, env);
}

class GripBorders : public GenericVideoFilter
{
public:
	GripBorders(PClip child, int w, int h, IScriptEnvironment* env);
	virtual PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);

private:
	PClip borders_clip;
};

GripBorders::GripBorders(PClip _child, int w, int h, IScriptEnvironment* env)
	: GenericVideoFilter(_child)
{
	int l = (w - vi.width) / 2, t = (h - vi.height) / 2,
		r = w - vi.width - l, b = h - vi.height - t;
	AVSValue args[] = { child, l, t, r, b };
	borders_clip = env->Invoke("AddBorders", AVSValue(args, 5)).AsClip();
	vi.width = w;
	vi.height = h;
}

PVideoFrame __stdcall GripBorders::GetFrame(int n, IScriptEnvironment* env)
{
	return borders_clip->GetFrame(n, env);
}

AVSValue __cdecl Create_GripBorders(AVSValue args, void* /* user_data */,
									IScriptEnvironment* env)
{
	enum ARGS { CLIP };

	AVSValue tw = env->GetVar("GripFit_target_width"),
		th = env->GetVar("GripFit_target_height"),
		rw = env->GetVar("GripFit_resize_width"),
		rh = env->GetVar("GripFit_resize_height");

	if(!th.Defined() || !rw.Defined() || !rh.Defined())
		env->ThrowError("GripBorders: Must run GripCrop before GripBorders");
	if(!tw.Defined())
		env->ThrowError("GripBorders: No target width was specified");

	const VideoInfo& vi = args[CLIP].AsClip()->GetVideoInfo();
	if(vi.width != rw.AsInt() || vi.height != rh.AsInt())
		env->ThrowError("GripBorders: Must run GripSize before GripBorders");

	return new GripBorders(args[CLIP].AsClip(), tw.AsInt(), th.AsInt(), env);
}

extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
	env->AddFunction("GripCrop", "c[width]i[height]i[source_anamorphic]b[dest_anamorphic]b"
		"[overscan]i[crop_round_width]i[crop_round_height]i[resize_round_width]i"
		"[resize_round_height]i[luma_threshold]i[samples]i", Create_GripCrop, 0);
	env->AddFunction("GripSize", "c[resizer]s", Create_GripSize, 0);
	env->AddFunction("GripBorders", "c", Create_GripBorders, 0);
	return "GripFit: Resize while preserving aspect ratio";
}
